(* non-embedded, functors, static dispatch *) 


use "../lib/smlprelude";




(* root objects -------------------------------------------------------	*)

open Closed;   (* closed datatype, % : closed *)

signature ROOT =
sig
    type 'a subself
    type self = closed subself
    type 'a ext = 'a subself
    val unpack : 'a subself -> ('a -> 'a subself) * 'a
    val pack : unit -> 'a -> 'a ext
    val new : unit -> self
end;

structure Root :> ROOT =
struct

(* types *)
type 'a subself = 'a
type 'a ext = 'a subself;
type self = closed subself;

(* packing *)
fun unpack ob = (I,ob);
fun pack () ext = I ext;

(* new *)
fun new () = pack () %;
end;



(* first position -----------------------------------------------------	*)

signature OBJECT =
sig
    type 'a ext
    type 'a subself
    type self = closed subself
    val unpack : 'a subself -> ('a -> 'a subself) * 'a
end;

signature PLACED =
sig
    structure Super : OBJECT
    type 'a ext
    type 'a subself = 'a ext Super.subself
    type self = closed subself
    val unpack : 'a subself -> ('a -> 'a subself) * 'a
    val pack : int -> 'a -> 'a ext
    val pos : 'a subself -> int
    val slide : 'a subself -> int -> 'a subself
    val eq : 'a subself -> 'b subself -> bool
    val dist : 'a subself -> 'b subself -> real
end;

functor  Place (structure S : OBJECT) :> PLACED where Super = S =
struct

structure Super = S;

(* types *)
type 'a ext = int * 'a;
type 'a subself = 'a ext Super.subself;
type self = closed subself;

(* packing *)
fun unpack p = explode (Super.unpack p);
fun repack f p = implode f (Super.unpack p);
fun pack n ext = (n, ext);

(* messages *)
fun pos p = let val (_,(x,_)) = Super.unpack p in x end;
fun slide p n =	repack (fn x => x+n) p;
fun eq p q = pos p = pos q;
fun dist p q = real (pos q - pos p);
end;


(* extra positions ----------------------------------------------------	*)

signature PLACEDOBJ =
sig
    type 'a ext
    type 'a subself
    type self = closed subself
    val unpack : 'a subself -> ('a -> 'a subself) * 'a
    val pos : 'a subself -> int
    val slide : 'a subself -> int -> 'a subself
    val eq : 'a subself -> 'b subself -> bool
    val dist : 'a subself -> 'b subself -> real
end;

functor  XPlace (structure S : PLACEDOBJ) :> PLACED where Super = S =
struct

structure Super = S;

(* types *)
type 'a ext = int * 'a;
type 'a subself = 'a ext Super.subself;
type self = closed subself;

(* packing *)
fun unpack p = explode (Super.unpack p);
fun repack f p = implode f (Super.unpack p);
fun pack n ext = (n, ext);

(* messages *)
fun pos p = let val (_,(x,_)) = Super.unpack p in x end;
fun slide p n =	repack (fn x => x+n) (Super.slide p n);
fun eq p q = Super.eq p q andalso pos p = pos q;
fun dist p q = sqrt (Real.+ (sq (Super.dist p q), sq (real (pos q - pos p))))
end;



(* color --------------------------------------------------------------	*)

signature PAINTED =
sig
    structure Super : OBJECT
    type 'a ext
    type 'a subself = 'a ext Super.subself
    type self = closed subself
    val pack : color -> 'a -> 'a ext
    val unpack : 'a subself -> ('a -> 'a subself) * 'a
    val col : 'a subself -> color
    val paint : 'a subself -> color -> 'a subself
    val eq : 'a subself -> 'b subself -> bool
end;

functor Paint (structure S : PLACEDOBJ) :> PAINTED where Super = S =
struct

structure Super = S;

(* types *)
type 'a ext = color * 'a;
type 'a subself = 'a ext Super.subself;
type self = closed subself;

(* packing *)
fun unpack p = explode (Super.unpack p);
fun repack f p = implode f (Super.unpack p);
fun pack c ext = (c, ext);

(* messages *)
fun col cp = let val (_,(c,_)) = Super.unpack cp in c end;
fun paint cp c = repack (fn _ => c) cp;
fun eq cp cq = col cp = col cq andalso Super.eq cp cq;
end;



(* line --------------------------------------------------------------- *)

signature MINIPOINT =
sig
    type 'a ext
    type 'a subself
    type self = closed subself
    val slide : 'a subself -> int -> 'a subself
    val eq : 'a subself -> 'b subself -> bool
    val dist : 'a subself -> 'b subself -> real
end;

signature LINE =
sig
    structure Param : MINIPOINT
    structure Super : OBJECT
    type ('a,'b) ext
    type ('a,'b) subself = ('a,'b) ext Super.subself
    type 'a self = ('a,closed) subself
    val unpack : ('a,'b) subself -> ('b -> ('a,'b) subself) * 'b
    val pack : ('a Param.subself * 'a Param.subself) -> 'b -> ('a,'b) ext
    val from : ('a,'b) subself -> 'a Param.subself
    val to : ('a,'b) subself -> 'a Param.subself
    val shift : ('a,'b) subself -> int -> ('a,'b) subself
    val eq : ('a,'b) subself -> ('c,'d) subself -> bool
    val length : ('a,'b) subself -> real
end;

functor Join (structure P : MINIPOINT and S : OBJECT) :> LINE
	where Super = S and Param = P =
struct

structure Param = P;
structure Super = S;

(* types *)
type ('a,'b) ext = ('a Param.subself * 'a Param.subself) * 'b;
type ('a,'b) subself = ('a,'b) ext Super.subself;
type 'a self = ('a,closed) subself;

(* packing *)
fun unpack p = explode (Super.unpack p);
fun repack f p = implode f (Super.unpack p);
fun pack ft ext = (ft, ext);

(* messages *)
fun from l = let val (_,((f,_),_)) = Super.unpack l in f end;
fun to l = let val (_,((_,t),_)) = Super.unpack l in t end;
fun shift l n =	repack (fn (f,t) => (Param.slide f n, Param.slide t n)) l;
fun eq p q = Param.eq (from p) (from q) andalso Param.eq (to p) (to q);
fun length l = Param.dist (from l) (to l);
end;




(* class hierarchy. ---------------------------------------------------- *)

structure Point =
    struct
	local
	structure P = Place (structure S = Root);
	in
	open P;
	fun pack n = Root.pack () o P.pack n;
	fun new n : self = pack n %;
	end;
    end;

structure Cpoint =
    struct
	local
	structure P = Paint (structure S = Point);
	in
	open P;
	fun pack (n,c) = Point.pack n o P.pack c;
	fun new (n,c) : self = pack (n,c) %;
	end;
    end;

structure D2point =
    struct
	local
	structure P = XPlace (structure S = Point);
	in
	open P;
	fun pack (x,y) = Point.pack x o P.pack y;
	fun new (x,y) : self = pack (x,y) %;
	end;
    end;

structure PointLine =
    struct
	local
	structure L = Join (structure S = Root and P = Point);
	in
	open L;
	fun pack ft = Root.pack () o L.pack ft;
	fun new ft : 'a self = pack ft %;
	end;
    end;

structure CpointLine =
    struct
	local
	(* slide and dist not specialized by Cpoint *)
	structure L = Join (structure S = Root 
			    and P = struct
					open Cpoint;
 					val slide = Point.slide;
					val dist = Point.dist;
				    end)
	in
	open L;
	fun pack ft = Root.pack () o L.pack ft;
	fun new ft : 'a self = pack ft %;
	end;
    end;

structure D2pointLine =
    struct
	local
	structure L = Join (structure S = Root and P = D2point);
	in
	open L;
	fun pack ft = Root.pack () o L.pack ft;
	fun new ft : 'a self = pack ft %;
	end;
    end;



