(* self-application, multiple dispatch, folding *)


use "../lib/smlprelude";



(* root objects -------------------------------------------------------	*)

open ExtensibleClosed;

signature ROOT =
sig
    type 'a subself
    type self = closed subself
    type wrap
    val pack   : unit -> ('a subself -> wrap) -> 'a -> 'a subself
    val unpack : 'a subself -> ('a -> 'a subself) * 'a
    val coerce : ('a subself -> 'b subself) ->
   	         ('b subself -> 'a subself) ->
                 'a subself -> 'b -> 'b subself
    exception Wrap
    val wrap'self : self -> wrap
    val self'wrap : wrap -> self
    val wrap : 'a subself -> wrap
    val new : unit -> self
end;

structure Root :> ROOT =
struct

(* types *)
datatype 'a state = S of
		'a subself -> wrap 			(* wrap   *)
withtype 'a subself = 'a state * 'a
and self = closed subself
and wrap = self;

(* wrapping *)
exception Wrap;
fun wrap'self (p : self) : wrap = p;
fun self'wrap (p : wrap) : self = p;
fun coerce c c' ((S w,_) :'a subself) e' : 'b subself = (S (w o c'),e');

(* messages *)
fun wrap (p as (S w,_)) = w p;
	
(* packing *)
fun unpack wob = explode (I,wob);
fun pack () wrap'curr ext = (S wrap'curr, ext);

(* new *)
fun new () = pack () wrap'self %;
end;



(* eqobjs -------------------------------------------------------------	*)

signature EQOBJ =
sig
    type 'a ext
    type 'a subself = 'a ext Root.subself
    type self = closed subself
    val pack   : unit -> ('a subself -> Root.wrap) -> 'a -> 'a subself
    val unpack : 'a subself -> ('a -> 'a subself) * 'a
    val repack : (('a subself -> Root.wrap -> bool) ->
		  ('a subself -> Root.wrap -> bool)) ->
		 'a subself -> 'a subself
    val coerce : ('a subself -> 'b subself) ->
		 ('b subself -> 'a subself) ->
                 'a subself -> 'b -> 'b subself
    val wrap'self : self -> Root.wrap
    val self'wrap : Root.wrap -> self
    exception Eq;
    val eq : 'a subself -> 'b subself -> bool
    val new : unit -> self
end;

structure Eqobj :> EQOBJ =
struct

(* types *)
datatype 'a state = S of
		'a subself -> Root.wrap -> bool		(* eq    *)
withtype 'a ext = 'a state * 'a
and 'a subself = 'a ext Root.subself;
type self = closed subself;

(* wrapping *)
val (Self,matches) = create (None : closed ext option);
fun super'self (p : self) : Root.self =
	let val (_,ext) = Root.unpack p
	in Root.coerce super'self self'super p (Self ext)
	end
and self'super (w : Root.self) : self =
	let val (_,ww) = Root.unpack w
	in case matches ww of
	  Some ext => Root.coerce self'super super'self w ext
	| _ => raise Root.Wrap
	end;
val wrap'self = Root.wrap'self o super'self;
val self'wrap = self'super o Root.self'wrap;
fun coerce c c' (p:'a subself) e' : 'b subself = 
	let val (_,(S q,e)) = Root.unpack p
	in Root.coerce c c' p (S (q o c'),e')
	end;

(* messages *)
exception Eq;
fun eq p q = let val (_,(S e,_)) = Root.unpack p in e p (Root.wrap q) end;

(* packing *)
fun unpack p = explode (Root.unpack p);
fun repack f p = implode (fn S m => S (f m)) (Root.unpack p);
fun oeq p q = raise Eq;
fun pack () wrap'curr ext = Root.pack () wrap'curr (S oeq,ext);

fun new () = pack () wrap'self %;
end;


(* points -------------------------------------------------------------	*)

signature POINT =
sig
    type 'a ext
    type 'a subself = 'a ext Eqobj.subself
    type self = closed subself
    val pack   : int -> ('a subself -> Root.wrap) -> 'a -> 'a subself
    val unpack : 'a subself -> ('a -> 'a subself) * 'a
    val repack : (('a subself -> int) *
		  ('a subself -> int -> 'a subself) *
		  ('a subself -> Root.wrap -> real) ->
		  ('a subself -> int) *
		  ('a subself -> int -> 'a subself) *
	          ('a subself -> Root.wrap -> real)) ->
	         'a subself -> 'a subself
    val coerce : ('a subself -> 'b subself) ->
		 ('b subself -> 'a subself) ->
                 'a subself -> 'b -> 'b subself
    val wrap'self : self -> Root.wrap
    val self'wrap : Root.wrap -> self
    val pos : 'a subself -> int
    val slide : 'a subself -> int -> 'a subself
    val dist : 'a subself -> 'b subself -> real
    val new : int -> self
end;

structure Point :> POINT =
struct

(* types *)
datatype 'a state = S of
		int *
		(('a subself -> int) *			(* pos	 *)
		 ('a subself -> int -> 'a subself) * 	(* slide *)
		 ('a subself -> Root.wrap -> real))	(* dist  *)
withtype 'a ext = 'a state * 'a
and 'a subself = 'a ext Eqobj.subself;
type self = closed subself;

(* wrapping *)
val (Self,matches) = create (None : closed ext option);
fun super'self (p : self) : Eqobj.self =
	let val (_,ext) = Eqobj.unpack p
	in Eqobj.coerce super'self self'super p (Self ext)
	end
and self'super (w : Eqobj.self) : self =
	let val (_,ww) = Eqobj.unpack w
	in case matches ww of
	  Some ext => Eqobj.coerce self'super super'self w ext
	| _ => raise Root.Wrap
	end;
val wrap'self = Eqobj.wrap'self o super'self;
val self'wrap = self'super o Eqobj.self'wrap;
fun coerce c c' (p:'a subself) e' : 'b subself = 
	let val (_,(S (x,(n,s,d)),e)) = Eqobj.unpack p
	in Eqobj.coerce c c' p (S (x,(n o c',fn n => c o (s o c') n,d o c')),e')
	end;

(* messages *)
fun pos p     = let val (_,(S (_,(n,_,_)),_)) = Eqobj.unpack p in n p end;
fun slide p n = let val (_,(S (_,(_,s,_)),_)) = Eqobj.unpack p in s p n end;
fun dist p q  = let val (_,(S (_,(_,_,d)),_)) = Eqobj.unpack p in d p (Root.wrap q) end;

(* packing *)
fun unpack p = explode (Eqobj.unpack p);
fun repack f p = implode (fn S (s,m) => S (s,f m)) (Eqobj.unpack p);
fun repacks f p = implode (fn S (s,m) => S (f s,m)) (Eqobj.unpack p);
fun mpos p = let val (_,(S (s,_),_)) = Eqobj.unpack p in s end;
fun mslide p n' = repacks (fn n => n + n') p;
fun mdist p q = real (pos (self'wrap q) - pos p);
fun peq eqeq p q = pos p = pos (self'wrap q) handle Root.Wrap => eqeq p q;
fun pack n wrap'curr ext =
	Eqobj.repack peq (Eqobj.pack () wrap'curr (S (n,(mpos, mslide, mdist)),ext));

(* new *)
fun new n = pack n wrap'self %;
end;



(* colored points  ----------------------------------------------------	*)

signature CPOINT =
sig
    type 'a ext
    type 'a subself = 'a ext Point.subself
    type self = closed subself
    val pack   : color * int -> ('a subself -> Root.wrap) -> 'a -> 'a subself
    val unpack : 'a subself -> ('a -> 'a subself) * 'a
    val repack : (('a subself -> color) *
		  ('a subself -> color -> 'a subself) ->
		  ('a subself -> color) *
		  ('a subself -> color -> 'a subself)) ->
	         'a subself -> 'a subself
    val coerce : ('a subself -> 'b subself) ->
		 ('b subself -> 'a subself) ->
                 'a subself -> 'b -> 'b subself
    val wrap'self : self -> Root.wrap
    val self'wrap : Root.wrap -> self
    val col : 'a subself -> color
    val paint : 'a subself -> color -> 'a subself
    val new : int * color -> self
end;

structure Cpoint :> CPOINT = 
struct

(* types *)
datatype 'a state = S of
		color *
		(('a subself -> color) *		(* col	 *)
		 ('a subself -> color -> 'a subself))	(* paint *)
withtype 'a ext = 'a state * 'a
and 'a subself = 'a ext Point.subself;
type self = closed subself;

(* wrapping *)
val (Self,matches) = create (None : closed ext option);
fun super'self (p : self) : Point.self =
	let val (_,ext) = Point.unpack p
	in Point.coerce super'self self'super p (Self ext)
	end
and self'super (w : Point.self) : self =
	let val (_,ww) = Point.unpack w
	in case (matches ww) of
	  Some ext => Point.coerce self'super super'self w ext
	| _ => raise Root.Wrap
	end;
val self'wrap = self'super o Point.self'wrap;
val wrap'self = Point.wrap'self o super'self;
fun coerce c c' (p:'a subself) e' : 'b subself = 
	let val (_,(S (x,(cl,pa)),e)) = Point.unpack p
	in Point.coerce c c' p (S (x,(cl o c',fn co => c o (pa o c') co)),e')
	end;

(* messages *)
fun col p     = let val (_,(S (_,(c,_)),_)) = Point.unpack p in c p end;
fun paint p c =	let val (_,(S (_,(_,t)),_)) = Point.unpack p in t p c end;

(* packing *)
fun unpack p = explode (Point.unpack p);
fun repack f p = implode (fn S (s,m) => S (s,f m)) (Point.unpack p);
fun repacks f p = implode (fn S (s,m) => S (f s,m)) (Point.unpack p);
fun mcol p = let val (_,(S (s,_),_)) = Point.unpack p in s end;
fun mpaint p c' = repacks (fn c => c') p;
fun cpeq peq p q =
	(col p = col (self'wrap q) andalso peq p q) handle Root.Wrap => peq p q;
fun pack (c,n) wrap'curr ext =
	Eqobj.repack cpeq (Point.pack n wrap'curr (S (c,(mcol,mpaint)),ext));

(* new *)
fun new (n,c) = pack (c,n) wrap'self %;
end;


(* d2point ----------------------------------------------------	*)

signature D2POINT =
sig
    type 'a ext
    type 'a subself = 'a ext Point.subself
    type self = closed subself
    val pack   : int * int -> ('a subself -> Root.wrap) -> 'a -> 'a subself
    val unpack : 'a subself -> ('a -> 'a subself) * 'a
    val repack : (('a subself -> int) ->
		  ('a subself -> int)) ->
	         'a subself -> 'a subself
    val coerce : ('a subself -> 'b subself) ->
		 ('b subself -> 'a subself) ->
                 'a subself -> 'b -> 'b subself
    val wrap'self : self -> Root.wrap
    val self'wrap : Root.wrap -> self
    val xpos : 'a subself -> int
    val ypos : 'a subself -> int
    val new : int * int -> self
end;

structure D2point :> D2POINT =
struct

(* types *)
datatype 'a state = S of
		int *
		('a subself -> int)		(* ypos  *)
withtype 'a ext = 'a state * 'a
and 'a subself = 'a ext Point.subself;
type self = closed subself;

(* wrapping *)
val (Self,matches) = create (None : closed ext option);
fun super'self (p : self) : Point.self =
	let val (_,ext) = Point.unpack p
	in Point.coerce super'self self'super p (Self ext)
	end
and self'super (w : Point.self) : self =
	let val (_,ww) = Point.unpack w
	in case (matches ww) of
	  Some ext => Point.coerce self'super super'self w ext
	| _ => raise Root.Wrap
	end;
val wrap'self = Point.wrap'self o super'self;
val self'wrap = self'super o Point.self'wrap;
fun coerce c c' (p:'a subself) e' : 'b subself = 
	let val (_,(S (x,po),e)) = Point.unpack p
	in Point.coerce c c' p (S (x,po o c'),e')
	end;

(* messages *)
val xpos = Point.pos; (* alias *)
fun ypos p = let val (_,(S (_,y),_)) = Point.unpack p in y p end;

(* packing *)
fun unpack p = explode (Point.unpack p);
fun repack f p = implode (fn S (s,m) => S (s,f m)) (Point.unpack p);
fun repacks f p = implode (fn S (s,m) => S (f s,m)) (Point.unpack p);
fun mypos p = let val (_,(S (s,_),_)) = Point.unpack p in s end;
fun d2pslide pslide p n = repacks (fn s => s+n) (pslide p n);
fun d2pdist pdist p q =
	sqrt (sq (real (ypos (self'wrap q) - ypos p)) + sq (pdist p q))
	handle Root.Wrap => pdist p q;
fun d2peq peq p q =
	(ypos p = ypos (self'wrap q) andalso peq p q) handle Root.Wrap => peq p q;
fun pack (x,y) wrap'curr ext =
	Eqobj.repack d2peq
		(Point.repack (fn (mp,sp,dp) =>	(mp,d2pslide sp,d2pdist dp))
			(Point.pack x wrap'curr (S (y,mypos),ext)));

(* new *)
fun new (x,y) = pack (x,y) wrap'self %;
end;




(* line -------------------------------------------------------	*)

signature MINIPOINT =
sig
    type 'a ext
    type 'a subself
    type self = closed subself
    val slide : 'a subself -> int -> 'a subself
    val dist : 'a subself -> 'b subself -> real
    val eq : 'a subself -> 'b subself -> bool
end;

signature LINE =
sig
    structure Param : MINIPOINT
    type 'a ext
    type 'a subself = 'a ext Eqobj.subself
    type self = closed subself
    val pack   : Param.self * Param.self -> ('a subself -> Root.wrap) ->
		 'a -> 'a subself
    val unpack : 'a subself -> ('a -> 'a subself) * 'a
    val repack : (('a subself -> Param.self) *
		  ('a subself -> Param.self) *
		  ('a subself -> int -> 'a subself) *
		  ('a subself -> real) ->
		  ('a subself -> Param.self) *
		  ('a subself -> Param.self) *
		  ('a subself -> int -> 'a subself) *
		  ('a subself -> real)) ->
		 'a subself -> 'a subself
    val wrap'self : self -> Root.wrap
    val self'wrap : Root.wrap -> self
    val coerce : ('a subself -> 'b subself) ->
		 ('b subself -> 'a subself) ->
                 'a subself -> 'b -> 'b subself
    val from : 'a subself -> Param.self
    val to : 'a subself -> Param.self
    val shift : 'a subself -> int -> 'a subself
    val length : 'a subself -> real
    val new : Param.self * Param.self -> self
end;

functor Line (structure P : MINIPOINT) :> LINE where Param = P =
struct

structure Param = P;

(* types *)
datatype 'a state = S of
		(Param.self * Param.self) *
		(('a subself -> Param.self) *		(* from *)
		 ('a subself -> Param.self) *		(* to   *)
		 ('a subself -> int -> 'a subself) *	(* shift *)
		 ('a subself -> real))			(* length *)
withtype 'a ext = 'a state * 'a
and 'a subself = 'a ext Eqobj.subself;
type self = closed subself;

(* wrapping *)  
val (Self,matches) = create (None : closed ext option);
fun super'self (l : self) : Eqobj.self =
	let val (_,ext) = Eqobj.unpack l
	in Eqobj.coerce super'self self'super l (Self ext)
	end
and self'super (w : Eqobj.self) : self =
	let val (_,ww) = Eqobj.unpack w
	in case (matches ww) of
	  Some ext => Eqobj.coerce self'super super'self w ext
	| _ => raise Root.Wrap
	end;
val self'wrap = self'super o Eqobj.self'wrap;
val wrap'self = Eqobj.wrap'self o super'self;
fun coerce c c' (l : 'a subself) e' : 'b subself = 
	let val (_,(S (x,(fr,to,sh,le)),e)) = Eqobj.unpack l
	in Eqobj.coerce c c' l (S (x,(fr o c',to o c',fn n => c o (sh o c') n,le o c')),e')
	end;

(* messages *)
fun from l    = let val (_,(S (_,(f,_,_,_)),_)) = Eqobj.unpack l in f l end;
fun to l      = let val (_,(S (_,(_,t,_,_)),_)) = Eqobj.unpack l in t l end;
fun shift l n = let val (_,(S (_,(_,_,s,_)),_)) = Eqobj.unpack l in s l n end;
fun length l  = let val (_,(S (_,(_,_,_,e)),_)) = Eqobj.unpack l in e l end;

(* packing *)
fun unpack p = explode (Eqobj.unpack p);
fun repack f p = implode (fn S (s,m) => S (s,f m)) (Eqobj.unpack p);
fun repacks f p = implode (fn S (s,m) => S (f s,m)) (Eqobj.unpack p);
fun state p = let val (_,(S (s,_),_)) = Eqobj.unpack p in s end;
fun mfr l = let val (f,_) = state l in f end;
fun mto l = let val (_,t) = state l in t end;
fun msh l n = repacks (fn (f,t) => (Param.slide f n, Param.slide t n)) l;
fun mlen l = Param.dist (from l) (to l);
fun leq eqeq p q =
	(Param.eq (from p) (from (self'wrap q)) andalso
	 Param.eq (to p) (to (self'wrap q)))
	handle Root.Wrap => eqeq p q;
fun pack (f,t) wrap'curr ext =
	Eqobj.repack leq
		(Eqobj.pack () wrap'curr (S ((f,t),(mfr,mto,msh,mlen)),ext));

(* new *)
fun new (f,t) = pack (f,t) wrap'self %;
end;


structure PointLine = Line (structure P = struct open Point; val eq = Eqobj.eq end);
structure CpointLine = Line (structure P = struct
						open Cpoint;
						val dist = Point.dist;
						val slide = Point.slide;
						val eq = Eqobj.eq
					   end);
(* etc, *)


